export class Board {

    constructor(width, height) {
        this.width = width
        this.height = height

        const columns = 'abcdefg'.split('')
            .reduce((accu, current) => {
                accu[current] = ''
                return accu
            }, {})

        // TODO Edge compability (Edge and Safari do not support Object Destructuring { ...[a, b, c] })
        //      Quite some ride to discover this... no syntax errors whatsoever
        //      Needed a very heavy binary search to find it (should be a bug in Edge i guess)
        //      replace with Object.assign https://caniuse.com/#search=spread :-(
        const rows = '7654321'.split('')
            .reduce((accu, current) => {
                // deep copy
                accu[current] = { ...columns }
                return accu
            }, {})
        this.rows = rows
    }

    init() {
        this.rows[7].a = 'o'
        this.rows[7].g = 'x'
        this.rows[1].a = 'x'
        this.rows[1].g = 'o'
        return this
    }

    stoneCount(board, color) {
        const predicate = cell => cell == color
        return board.flat().filter(predicate).length
    }

    // compare to a next or previous board (bail out if impossible moves, i.e. not succeeding boards are past)
    // does not know if the board passed is previous or next (because that cannot be derived from the move if it was not a clone or an infect)
    // This can be used for retrieving a move (which is not available if the move was something like 'a6', see discussion about ambiguous moves) (see long.pgn for an example, move 7 )
    // In fact we only need to pass the color and single position as well, to find out
    // where we came from.
    diff(other) {
        const diff = []

        for (const [rowIndex, row] of Object.entries(this.rows)) {
            for (const [columnIndex, cell] of Object.entries(row)) {
                if (cell != other[rowIndex][columnIndex]) {
                    if (cell != '') {
                        diff.push({ board: 0, rowIndex, columnIndex })
                    } else if (other[rowIndex][columnIndex] != '') {
                        diff.push({ board: 1, rowIndex, columnIndex })
                    }
                }
            }
        }

        diff.sort((a, b) => b.board - a.board)

        return diff
    }

    // move piece from e3 to f2 and diff the boards to get the move (e3f2)
    getMoveFromDiff(diff) {
        const a = new Board()
        a.init()
        a.rows[3].e = 'x'       // e3
        console.table(a.rows)

        const b = new Board()
        b.init()
        b.rows[2].f = 'x'       // f2
        console.table(b.rows)

        console.table(b.diff(a.rows))
    }

    // parse single FEN string (e.g. xx3oo/6o/7/7/7/7/o4xx)
    static fromFen(fenString) {
        // TODO replace with files and ranks
        const board = Array.from(Array(7), () => new Array(7).fill(''))

        let [row, column] = [0, 0]
        for (const character of fenString) {
            switch (character) {
                case 'o': case 'x':
                    board[row][column] = character
                    column++
                    break

                case '/':
                    row++
                    column = 0
                    break

                default:
                    // run length
                    if (+character >= 1 && +character <= 7) {
                        column += +character
                    } else {
                        console.warn(`'invalid FEN character ${character}`)
                    }
            }
        }

        if (row * 7 + column > 7 * 7) {
            console.error(`FEN overflow: '${fenString}', cell count: ${row * 7 + column}`)
        }

        // TODO return a new Board object
        return board
    }
}
