// easing functions

const bezier = t => t * t * (3.0 - 2.0 * t)
const lerp = t => t
const easeIn = t => t = t * t
const easeOut = t => t = t * (2 - t)
const easyInOut = t => t < .5 ? 2 * t * t : -1 + (4 - 2 * t) * t

const tween = (v0, v1, t) => (1 - easyInOut(t)) * v0 + easyInOut(t) * v1

export class CanvasRenderer {

    constructor($canvas) {
        const context = $canvas.getContext('2d')
        const [width, height] = [$canvas.width, $canvas.height]

        if (width % 7 != 0 || height % 7 != 0) {
            console.warn(`[CanvasRenderer] crispy rendering: dimensions ${width}*${height} are not a multiple of 7`)
        }
        const [WIDTH, HEIGHT] = [7, 7]

        this.CELLSIZE = canvas.height / WIDTH

        const epsilon = .01

        const cellSize = width / WIDTH

        board.addEventListener('move', e => {
            console.log({ e })      // type, source, destination
            switch (e.type) {
                case 'clone':

                case 'infect':

                case 'jump':
            }
        })

        asyncAnimate()
    }

    drawPiece(column, row, color, size) {
        ctx.beginPath()
        ctx.fillStyle = color
        ctx.arc(
            cellSize * (column + .5), cellSize * (row + .5),
            size,
            0, 2 * Math.PI
        )
        ctx.fill()
    }

    drawBoard() {
        for (let row = 0; row < HEIGHT; row++) {
            for (let column = 0; column < WIDTH; column++) {
                ctx.fillStyle = row % 2 ^ column % 2 ? 'lightgray' : 'gray'
                ctx.fillRect(column * cellSize, row * cellSize, cellSize, cellSize)
            }
        }
    }


    clone(source, destination, time) {
        // clone (one step): stay, birth (small disc) -> grow and move
        const x = tween(source.column, destination.column, time)
        const y = tween(source.row, destination.row, time)

        drawBoard()
        drawPiece(0, 0, 'white', cellSize / 3)
        drawPiece(1, 2, 'black', cellSize / 3)
        drawPiece(2, 2, 'black', cellSize / 3)

        drawPiece(x, y, 'white', cellSize * time / 3)

        // if(time <= 1.0) {
        //     requestAnimationFrame(() => clone(start, end, time + .025))
        // }
    }

    jump(source, destination, time) {
        // jump (two steps): krimp -> langgerekt -> jump -> grow

    }

    infect(source, destination, time) {
        // after clone or jump: infect (neighbours of enemy) 
        // like clone but replace enemies piece 
        // clone (one step): stay, birth (small disc) -> grow and move
        const x = tween(source.column, destination.column, time)
        const y = tween(source.row, destination.row, time)

        drawBoard()

        // draw pieces
        drawPiece(0, 0, 'white', cellSize / 3)
        drawPiece(1, 1, 'white', cellSize / 3)
        drawPiece(1, 2, 'black', cellSize / 3)
        drawPiece(2, 2, 'black', cellSize / 3)

        drawPiece(x, y, 'white', cellSize * time / 3)

        // arrow (needed?)
        const radians = -Math.atan2(parentPos.x - childPos.x, parentPos.y - childPos.y)

        context.save()
        context.translate(childPos.x, childPos.y)
        context.rotate(radians)

        context.moveTo(25, 0)
        context.lineTo(0, 75)
        context.lineTo(-25, 0)

        context.restore()
    }

    async asyncAnimate() {
        // very nice async + raf -> animation in a loop (instead of ticking)
        const start = { row: 0, column: 0 }
        const end = { row: 1, column: 1 }
        for (let time = 0; time <= 1 + epsilon; time += .025) {
            await new Promise(resolve => {
                requestAnimationFrame(resolve);
                clone(start, end, time)
            })
        }

        // infect first black stone (getNeighours -> findEnemies)
        console.log('infect')
        start = { row: 1, column: 1 }
        end = { row: 2, column: 1 }
        for (let time = 0; time <= 1 + epsilon; time += .025) {
            await new Promise(resolve => {
                requestAnimationFrame(resolve);
                infect(start, end, time)
            })
        }

        // infect second black stone
        console.log('first')
        start = { row: 1, column: 1 }
        end = { row: 2, column: 2 }
        for (let time = 0; time <= 1 + epsilon; time += .025) {
            await new Promise(resolve => {
                requestAnimationFrame(resolve);
                infect(start, end, time)
            })
        }

    }
}
